#!/bin/sh
# Copyright 2015 National Instruments. All Rights Reserved

niecatwarn_canNotBackToEthernet="680480"
niecatwarn_unclaimFailedNotClaimedBefore="680482"
niecatwarn_claimFailedAlreadyClaimed="680481"
niecatwarn_toolAlreadyInited="680483"
niecaterr_changeModeFailed="-680487"
niecaterr_claimFailedAlreadyClaimedByOther="-680482"
niecaterr_canNotChangeModeWhenClaimed="-680480"
niecaterr_toolNotInited="-680491"
niecaterr_toolAlreadyLocked="-680490"
niecaterr_toolAlreadyInited="-680489"
niecaterr_masterNotFound="-680486"
niecaterr_invalidModeParameter="-680488"
niecaterr_invalidMasterIdParameter="-680483"
niecaterr_invalidInitParameter="-680481"
niecaterr_invalidPidParameter="-680485"
niecaterr_unclaimFailedAlreadyClaimedByOther="-680484"

declare -A kEcatErrorMap
kEcatErrorMap=(
[$niecatwarn_canNotBackToEthernet]="[$niecatwarn_canNotBackToEthernet] Internal error. Contact NI for help."
[$niecatwarn_unclaimFailedNotClaimedBefore]="[$niecatwarn_unclaimFailedNotClaimedBefore] Failed to unclaim the network adapter for EtherCAT. This adapter has not been claimed before."
[$niecatwarn_claimFailedAlreadyClaimed]="[$niecatwarn_claimFailedAlreadyClaimed] Failed to claim the network adapter for EtherCAT. This adapter has been claimed before."
[$niecatwarn_toolAlreadyInited]="[$niecatwarn_toolAlreadyInited] NI EtherCAT Tool has already been initiated."
[$niecaterr_changeModeFailed]="[$niecaterr_changeModeFailed] Failed to execute the change mode command on the network adapter. Make sure you have the permission or the installation is correct."
[$niecaterr_claimFailedAlreadyClaimedByOther]="[$niecaterr_claimFailedAlreadyClaimedByOther] Failed to claim the network adapter for EtherCAT. The adapter has been claimed by another application."
[$niecaterr_canNotChangeModeWhenClaimed]="[$niecaterr_canNotChangeModeWhenClaimed] Failed to change the mode of the network adapter for EtherCAT. The adapter has been claimed by another application."
[$niecaterr_toolNotInited]="[$niecaterr_toolNotInited] The EtherCAT Tool has not been initiated. Run the init command first."
[$niecaterr_toolAlreadyLocked]="[$niecaterr_toolAlreadyLocked] The EtherCAT Tool has been locked for a long time. There is another process accessing this script exclusively."
[$niecaterr_toolAlreadyInited]="[$niecaterr_toolAlreadyInited] The EtherCAT Tool has already been initiated. You cannot initialize the tool twice."
[$niecaterr_masterNotFound]="[$niecaterr_masterNotFound] This master cannot be found. Please use -list command to get the available masters."
[$niecaterr_invalidModeParameter]="[$niecaterr_invalidModeParameter] The EtherCAT tool received invalid mode parameters. Run niecattool -h for help."
[$niecaterr_invalidMasterIdParameter]="[$niecaterr_invalidMasterIdParameter] The EtherCAT tool received invalid master id parameters. Run niecattool -h for help."
[$niecaterr_invalidInitParameter]="[$niecaterr_invalidInitParameter] The EtherCAT tool received invalid init parameters. Run niecattool -h for help."
[$niecaterr_invalidPidParameter]="[$niecaterr_invalidPidParameter] The Ethercat tool received invalid pid parameters. Run niecattool -h command for help."
[$niecaterr_unclaimFailedAlreadyClaimedByOther]="[$niecaterr_unclaimFailedAlreadyClaimedByOther] This network adapter has been claimed for EtherCAT by another application."
)


NATI_PATH=/tmp/natinst
ECAT_PATH=$NATI_PATH/niecat
LIST_PATH=$ECAT_PATH/list

_driver_type=""
_get_driver_type()
{
   local id=$1
   _driver_type=`awk -F "[=|,]" '{if ($2 ~ /^'"$id"'$/) print $6}' $LIST_PATH`
}

_bus_info=""
_get_bus_info()
{
   local id=$1
   _bus_info=`awk -F "[=|,|;]" '{if ($2 ~ /^'"$id"'$/) print $12}' $LIST_PATH`
}

_ifr_name=""
_get_ifr_name_from_id()
{
   local id=$1
   _ifr_name=`awk -F "[=|,|;]" '{if ($2 ~ /^'"$id"'$/) print $10}' $LIST_PATH`
}

DEV_PATH=$ECAT_PATH/dev
ECAT=ecat

_current_mode=""
_get_current_mode()
{
   local id=$1
   _current_mode="ethernet"
   if [ -d "$DEV_PATH/$ECAT$id" ];then
      _current_mode="ethercat"
   fi
}

_original_mode=""
_get_original_mode()
{
   local id=$1
   local pid=$2
   _original_mode=`ls $DEV_PATH/$ECAT$id/$pid 2>/dev/null`
}

_owner_pid=""
_get_owner_pid()
{
   local id=$1
   _owner_pid=""
   local mode_path=$DEV_PATH/$ECAT$id
   if [ -d "$mode_path" ]; then
      _owner_pid=`ls $mode_path 2>/dev/null`
   fi
}

_quiet_mode=""
_print()
{
   local error=$1
   if [ "$_quiet_mode" == "quiet" ];then
      return
   else
      echo $error
   fi
}

_bind_driver()
{
   local id=$1
   local driver_path=$2
   _get_bus_info $id
   (echo $_bus_info > $driver_path/bind) >/dev/null 2>&1
   if [ $? -ne 0 ]; then
      _print "${kEcatErrorMap[$niecaterr_changeModeFailed]}"
      exit 1
   fi
}

_unbind_driver()
{
   local id=$1
   local driver_path=$2
   _get_bus_info $id
   (echo $_bus_info > $driver_path/unbind) >/dev/null 2>&1
   if [ $? -ne 0 ]; then
      _print "${kEcatErrorMap[$niecaterr_changeModeFailed]}"
      exit 1
   fi
}

IGB_DRIVER_PATH=/sys/bus/pci/drivers/igb
MACB_DRIVER_PATH=/sys/bus/platform/drivers/macb
NIEC_DRIVER_PATH=/sys/bus/platform/drivers/ni_ethercat

_igb_change_to_ethercat()
{
   local id=$1
   _unbind_driver $id $IGB_DRIVER_PATH
}

_change_to_ethercat()
{
   local id=$1
   _get_driver_type $id
   if [ "$_driver_type" == "macb" ]; then
      _unbind_driver $id $MACB_DRIVER_PATH
      _bind_driver $id $NIEC_DRIVER_PATH

      _get_bus_info $id
      ln -sf "/dev/niecat$_bus_info" "/dev/niecat$id"
   else
      _igb_change_to_ethercat $id
   fi

   mkdir $DEV_PATH/$ECAT$id
}

_igb_change_to_ethernet()
{
   local id=$1
   _bind_driver $id $IGB_DRIVER_PATH
}

_macb_change_to_ethernet()
{
   _unbind_driver $1 $NIEC_DRIVER_PATH
   _bind_driver $1 $MACB_DRIVER_PATH
}

_change_to_ethernet()
{
   local id=$1
   _get_driver_type $id
   if [ "$_driver_type" == "macb" ]; then
      _macb_change_to_ethernet $id
   else
      _igb_change_to_ethernet $id
   fi
   rm -rf $DEV_PATH/$ECAT$id
   _get_ifr_name_from_id $1
   ifconfig $_ifr_name up
}

_clean_owner()
{
   local id=$1
   local pid=$2
   _get_original_mode $id $pid
   if [ "$_original_mode" == "ethernet" ];then
      _change_to_ethernet $id
   fi
   rm -rf $DEV_PATH/$ECAT$id/$pid
}

_purge_dangling_masters()
{
   local masters=`ls $DEV_PATH 2>/dev/null`
   for master in $masters
   do
      local pid=`ls $DEV_PATH/$master 2>/dev/null`
      if [ -z "$pid" ];then
         continue
      fi
      local alive=`ps -ef | awk -v p="$pid" '{if($1==p) print $1}'`
      if [ -z "$alive" ];then
         local id=${master:${#ECAT}}
         _clean_owner $id $pid
      fi
   done
}

_check_init()
{
   if [ ! -f $LIST_PATH ]; then
      _print "${kEcatErrorMap[$niecaterr_toolNotInited]}"
      exit 1
   fi
}
_check_master_id()
{
   local id=$1
   if ! [[ $id =~ ^[0-9]+$ ]]; then
      _print "${kEcatErrorMap[$niecaterr_invalidMasterIdParameter]}"
      exit 1
   fi

   local dev=`awk -F "[=|,]" '{if ($2 ~ /^'"$id"'$/) print $2}' $LIST_PATH`
   if [ -z "$dev" ];then
      _print "${kEcatErrorMap[$niecaterr_masterNotFound]}"
      exit 1
   fi
}

_check_pid()
{
   local pid=$1
   if ! [[ $pid =~ ^[0-9]+$ ]]; then
      _print "${kEcatErrorMap[$niecaterr_invalidPidParameter]}"
      exit 1
   fi
}

claim()
{
   _check_init
   _purge_dangling_masters

   local id=$1
   local pid=$2

   _check_master_id $id
   _check_pid $pid

   _get_owner_pid $id
   if [ -n "$_owner_pid" ];then
      if [ "$_owner_pid" != "$pid" ];then
         _print "${kEcatErrorMap[$niecaterr_claimFailedAlreadyClaimedByOther]}"
         exit 1
      else
         _print "${kEcatErrorMap[$niecatwarn_claimFailedAlreadyClaimed]}"
         exit 0
      fi
   fi

   _get_current_mode $id
   if [ "$_current_mode" == "ethernet" ];then
      _change_to_ethercat $id
   fi
   local master_path=$DEV_PATH/$ECAT$id
   mkdir -p $master_path/$pid
   touch $master_path/$pid/$_current_mode
}

unclaim()
{
   _check_init
   _purge_dangling_masters

   local id=$1
   local pid=$2

   _check_master_id $id
   _check_pid $pid

   _get_owner_pid $id
   if [ -z "$_owner_pid" ];then
      _print "${kEcatErrorMap[$niecatwarn_unclaimFailedNotClaimedBefore]}"
      exit 0
   fi
   if [ "$_owner_pid" != "$pid" ];then
      _print "${kEcatErrorMap[$niecaterr_unclaimFailedAlreadyClaimedByOther]}"
      exit 1
   fi

   _clean_owner $id $pid
}

list()
{
   _check_init
   cat $LIST_PATH | awk 'gsub(/,busInfo=[^;]*/, "")'
}

mode()
{
   _check_init
   _purge_dangling_masters

   local id=$1
   local target_mode=$2

   _check_master_id $id
   _get_owner_pid $id
   if [ -n "$_owner_pid" ]; then
      _print "${kEcatErrorMap[$niecaterr_canNotChangeModeWhenClaimed]}"
      exit 1
   fi

   _get_current_mode $id
   if [ "$target_mode" == "ethercat" ];then
      if [ "$_current_mode" == "ethernet" ];then
         _change_to_ethercat $id
      fi
   elif [ "$target_mode" == "ethernet" ];then
      if [ "$_current_mode" == "ethercat" ];then
         _change_to_ethernet $id
      fi
   else
      _print "${kEcatErrorMap[$niecaterr_invalidModeParameter]}"
      exit 1
   fi
}

_igb_change_mode()
{
   local dev=$1
}

init()
{
   local force=$1
   if [ -d "$ECAT_PATH" ];then
      if [ "$force" == "force" ];then
         _print "${kEcatErrorMap[$niecatwarn_toolAlreadyInited]}"
         rm -rf "$ECAT_PATH"
      elif [ -z "$force" ];then
         _print "${kEcatErrorMap[$niecaterr_toolAlreadyInited]}"
         exit 1
      else
         _print "${kEcatErrorMap[$niecaterr_invalidInitParameter]}"
         exit 1
      fi
   fi
   local ls_dev=`ls /sys/class/net`
   local devs=(${ls_dev})
   local count=0
   local output=""
   for dev in ${devs[@]}
   do
      local type=`(ethtool -i ${dev}) 2> /dev/null | awk '/driver/{print $NF}'`
      local bus_info=`(ethtool -i ${dev}) 2> /dev/null | awk '/bus-info/{print $NF}'`
      if [ "$type" == "igb" ];then
         local link_id=`echo $bus_info |
            awk -F '[.|:]' '{printf("0x01%2s%2s%2s",$2,$3,$4);}' |
            tr ' ' '0'`
         _igb_change_mode $dev
         
      elif [ "$type" == "macb" ];then
         local link_id="0x"`expr $count + 1`
      else
         continue
      fi
      local mac=`cat /sys/class/net/$dev/address`
      local id=`expr $count`
      output+="masterId="$id",mac="$mac",type="$type",linkId="$link_id",ifrName="$dev",busInfo="$bus_info";\n"
      ((++count))
   done

   mkdir -p "$DEV_PATH"
   echo -e ${output%\\n*} > $LIST_PATH
}

command()
{
   case "$1" in
      -claim) claim $2 $3;;
      -unclaim) unclaim $2 $3;;
      -list) list;;
      -mode) mode $2 $3;;
      -init) init $2;;
      -help)
         echo "niecattool version 1.0.0"
         echo "Usage:"
         echo "   -claim [quiet] MASTERID PID    Process PID claims the adapter of MASTERID"
         echo "   -unclaim [quiet] MASTERID PID  Process PID unclaims the adapter of MASTERID"
         echo "   -list [quiet]                  List the adapters available for EtherCAT"
         echo "   -mode [quiet] MASTERID ethercat|ethernet  Change the adapter of MASTERID to ethercat|ethernet"
         echo "   -init [quiet][force]           Init the script tool"
         echo "   -help                          Show this help"
         ;;
      *)
       echo "Usage: {-claim|-unclaim|-list|-mode|-init}"
       echo "For mode information run niecattool -help"
       exit 1
       ;;
   esac
}

LOCK_PATH=$NATI_PATH/niecat.lock
(
   for i in `seq 5`
   do
      flock -xn 200
      if [ $? -ne 0 ]; then
         sleep 1
      else
         if [ "$2" == "quiet" ]; then
            _quiet_mode="quiet"
            command $1 $3 $4
         else
            command $1 $2 $3
         fi
         exit
      fi
   done
   _print "${kEcatErrorMap[$niecaterr_toolAlreadyLocked]}"
   exit 1
) 200>$LOCK_PATH
